#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from artan.state.stateful_transformer import StatefulTransformer


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from artan.state import StatefulTransformer
from artan.filter.filter_params import (
    KalmanFilterParams, HasMultipleModelAdaptiveEstimationEnabled,
    HasMultipleModelMeasurementWindowDuration)


class LinearKalmanFilter(StatefulTransformer, KalmanFilterParams, HasMultipleModelAdaptiveEstimationEnabled,
                         HasMultipleModelMeasurementWindowDuration):
    """
    Linear Kalman Filter, implemented with a stateful spark Transformer for running parallel filters /w spark
    dataframes. Transforms an input dataframe of noisy measurements to dataframe of state estimates using stateful
    spark transformations, which can be used in both streaming and batch applications.

    Assuming a state vector :math:`x_k` with size `stateSize`, and measurements vector :math:`z_k`
    with size `measurementSize`, below parameters can be specified.

    - :math:`F_k`, process model, matrix with dimensions `stateSize` x `stateSize`
    - :math:`H_k`, measurement model, matrix with dimensions `stateSize` x `measurementSize`
    - :math:`Q_k`, process noise covariance, matrix with dimensions `stateSize` x `stateSize`
    - :math:`R_k`, measurement noise covariance, matrix with dimensions `measurementSize` x `measurementSize`
    - :math:`u_k`, optional control vector, vector with size `controlSize`
    - :math:`B_k`, optional control model, matrix with dimensions `stateSize` x `controlSize`

    Linear Kalman Filter will predict & estimate the state according to following state and measurement equations.

    .. math::

        x_k &= F_k x_{k-1} + B_k u_k + v_k \\

        z_k &= H_k x_k + w_k

    Where :math:`v_k` and :math:`w_k` are noise vectors drawn from zero mean,
    :math:`Q_k` and :math:`R_k` covariance distributions.

    The default values of system matrices will not give you a functioning filter, but they will be initialized
    with reasonable values given the state and measurement sizes. All of the inputs to the filter can
    be specified with a dataframe column which will allow you to have different value across measurements/filters,
    or you can specify a constant value across all measurements/filters.
    """
    def __init__(self, stateSize, measurementSize):
        super(LinearKalmanFilter, self).__init__()
        self._java_obj = self._new_java_obj("com.github.ozancicek.artan.ml.filter.LinearKalmanFilter",
                                            stateSize, measurementSize, self.uid)


    def setMultipleModelMeasurementWindowDuration(self, value):
        """
        Optionally set the window duration for grouping measurements in same window for MMAE filter aggregation.
        Could be used for limiting the state on streaming if event time column is set.

        :param value: String
        :return: LinearKalmanFilter
        """
        return self._set(multipleModelMeasurementWindowDuration=value)

    def setEnableMultipleModelAdaptiveEstimation(self):
        """
        Enable MMAE output mode

        :return: LinearKalmanFilter
        """
        return self._set(multipleModelAdaptiveEstimationEnabled=True)


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#


from pyspark.sql.column import (
    Column, _to_java_column, _to_seq
)
from pyspark import SparkContext
import pyspark.sql.functions as F
import numpy as np


def _sc():
    return SparkContext._active_spark_context


def _spark_functions():
    return _sc()._jvm.com.github.ozancicek.artan.ml.SparkFunctions


def _function_factory(cols, func):
    return Column(func.apply(_to_seq(_sc(), cols, _to_java_column)))


def arrayToVector(arrayCol):
    """
    Converts array of doubles column to dense vector column

    :param arrayCol: Array of doubles column, either string column name or column expression
    :return: DenseVector column
    """
    return _function_factory([arrayCol], _spark_functions().arrayToVector())


def vectorToArray(vectorCol):
    """
    Converts vector column to array of doubles

    :param vectorCol: Vector column, either string column name or column expression
    :return: Array of doubles column
    """
    return _function_factory([vectorCol], _spark_functions().vectorToArray())


def arrayToMatrix(numRowsCol, numColsCol, arrayCol):
    """
    Creates dense matrix from array of doubles column

    :param numRowsCol: Number of rows column, either integer or column expression
    :param numColsCol: Number of cols column, either integer or column expression
    :param arrayCol: Array of doubles column expression, should be column major ordered
    :return: DenseMatrix column
    """
    numRows = F.lit(numRowsCol) if isinstance(numRowsCol, int) else numRowsCol
    numCols = F.lit(numColsCol) if isinstance(numColsCol, int) else numColsCol
    return _function_factory([numRows, numCols, arrayCol], _spark_functions().arrayToMatrix())


def matrixToArray(matrixCol):
    """
    Converts matrix column to array of doubles

    :param matrixCol: Matrix column, either string column name or column expression
    :return: struct of (numRows, numCols, values) where values are column major values of matrix
    """
    return _function_factory([matrixCol], _spark_functions().matrixToArray())



def zerosVector(sizeCol):
    """
    Creates vector of zeros column

    :param sizeCol: Size of the vector, either integer or column expression
    :return: Vector column
    """
    size = F.lit(sizeCol) if isinstance(sizeCol, int) else sizeCol
    return _function_factory([size], _spark_functions().zerosVector())


def onesVector(sizeCol):
    """
    Creates vector of ones column

    :param sizeCol: Size of the vector, either integer or column expression
    :return: Vector column
    """
    size = F.lit(sizeCol) if isinstance(sizeCol, int) else sizeCol
    return _function_factory([size], _spark_functions().onesVector())


def dotVector(leftVectorCol, rightVectorCol):
    """
    Returns the dot product of two vector columns

    :param leftVectorCol: Vector column
    :param rightVectorCol: Vector column
    :return: Double column
    """
    return _function_factory([leftVectorCol, rightVectorCol], _spark_functions().dotVector())


def scalVector(scaleCol, vectorCol):
    """
    Scales the vector column with a scalar constant

    :param scaleCol: scale constant, either numeric or column expression
    :param vectorCol: Vector column
    :return: Vector column
    """
    scale = F.lit(float(scaleCol)) if isinstance(scaleCol, (int, float)) else scaleCol
    return _function_factory([scale, vectorCol], _spark_functions().scalVector())


def axpyVector(alphaCol, xVectorCol, yVectorCol):
    """
    Axpy operation on vectors, alpha*x +y where alpha is scalar, x and y vectors

    :param alphaCol: scalar which scales x, either numeric or column expression
    :param xVectorCol: Vector column corresponding to x
    :param yVectorCol: Vector column corresponding to y
    :return: Vector column
    """
    alpha = F.lit(float(alphaCol)) if isinstance(alphaCol, (int, float)) else alphaCol
    return _function_factory([alpha, xVectorCol, yVectorCol], _spark_functions().axpyVector())


def eyeMatrix(sizeCol):
    """
    Creates identity matrix column

    :param sizeCol: Size of the matrix, either integer or column expression
    :return: DenseMatrix column
    """
    size = F.lit(sizeCol) if isinstance(sizeCol, int) else sizeCol
    return _function_factory([size], _spark_functions().eyeMatrix())


def zerosMatrix(numRowsCol, numColsCol):
    """
    Creates zeros matrix column

    :param numRowsCol: Number of rows column, either integer or column expression
    :param numColsCol: Number of cols column, either integer or column expression
    :return: Matrix column
    """
    numRows = F.lit(numRowsCol) if isinstance(numRowsCol, int) else numRowsCol
    numCols = F.lit(numColsCol) if isinstance(numColsCol, int) else numColsCol
    return _function_factory([numRows, numCols], _spark_functions().zerosMatrix())


def diagMatrix(diagVectorCol):
    """
    Creates a diag matrix from vector

    :param diagVectorCol: Vector column corresponding to diagonal
    :return: Matrix column
    """
    return _function_factory([diagVectorCol], _spark_functions().diagMatrix())


def randnMultiGaussian(meanArray, covMatrix, seed=0):
    """
    Samples from multivariate gaussian as vector

    :param meanArray: mean of the distribution, either List[Float] or numpy array
    :param covMatrix: covariance of the distribution, either List[List[Float]]] (row major) or numpy 2d array
    :param seed: seed of the rand
    :return: DenseVector column
    """
    root = np.linalg.cholesky(np.array(covMatrix))
    rows, columns = root.shape

    root = arrayToMatrix(
        F.lit(rows), F.lit(columns),
        F.array([F.lit(el) for el in root.reshape(int(rows*columns), order="F").tolist()]))
    mean = arrayToVector(F.array([F.lit(float(el)) for el in meanArray]))

    samples = arrayToVector(F.array([F.randn(seed=seed+el) for el in range(0, len(meanArray))]))
    return _function_factory([mean, root, samples], _spark_functions().scaleToMultiGaussian())


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import numpy as np

from pyspark.ml.param import Params, Param, TypeConverters
from pyspark.ml.param.shared import HasLabelCol, HasFeaturesCol
from pyspark.ml.linalg import DenseMatrix
from pyspark.ml.common import inherit_doc
from artan.state import StatefulTransformer
from artan.filter.filter_params import HasInitialState, HasInitialStateCol


class _HasForgettingFactor(Params):
    """
    Mixin for param forgetting factor, between 0 and 1.
    """

    forgettingFactor = Param(
        Params._dummy(),
        "forgettingFactor", "Forgetting factor between 0 and 1", typeConverter=TypeConverters.toFloat)

    def __init__(self):
        super(_HasForgettingFactor, self).__init__()

    def getForgettingFactor(self):
        """
        Gets the value of forgettingFactor or its default value.
        """
        return self.getOrDefault(self.forgettingFactor)


class _HasRegularizationMatrix(Params):
    """
    Mixin for param regularization matrix.
    """

    regularizationMatrix = Param(
        Params._dummy(),
        "regularizationMatrix",
        "Positive definite regularization matrix for RLS filter, typically a factor multiplied by identity matrix." +
        "Small factors (>1) give more weight to initial state, whereas large factors (>>1) decrease regularization" +
        "and cause RLS filter to behave like ordinary least squares",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(_HasRegularizationMatrix, self).__init__()

    def getRegularizationMatrix(self):
        """
        Gets the value of regularization matrix or its default value.
        """
        return self.getOrDefault(self.regularizationMatrix)


class _HasRegularizationMatrixCol(Params):
    """
    Mixin for param regularization matrix column.
    """

    regularizationMatrixCol = Param(
        Params._dummy(),
        "regularizationMatrixCol",
        "Regularization matrix column for specifying different reg matrices across filters",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(_HasRegularizationMatrixCol, self).__init__()

    def getRegularizationMatrixCol(self):
        """
        Gets the value of regularization matrix column or its default value.
        """
        return self.getOrDefault(self.regularizationMatrixCol)


@inherit_doc
class RecursiveLeastSquaresFilter(StatefulTransformer, HasInitialState, HasInitialStateCol,
                                  _HasForgettingFactor, _HasRegularizationMatrix, _HasRegularizationMatrixCol,
                                  HasLabelCol, HasFeaturesCol):
    """
    Recursive formulation of least squares with exponential weighting & regularization, implemented with
    a stateful spark Transformer for running parallel filters /w spark dataframes. Transforms an input dataframe
    of observations to a dataframe of model parameters using stateful spark transformations, which can be used
    in both streaming and batch applications.

    Let :math:`w` denote the model parameters and :math:`w_{est}` denote our prior belief. RLS minimizes following
    regularization and weighted SSE terms;

    .. math::

        (w - w_{est}) \cdot (\lambda^{-N-1}P)^{-1} \cdot (w - w_{est}) + \sum_{k=1}^{N} \lambda^{N - k} (d_k - u_k \cdot w)

    Where:

    - :math:`\lambda`: forgetting factor, or exponential weighting factor. Between 0 and 1.
    - :math:`P`: regularization matrix. Smaller values increseas the weight of regularization term, whereas larger
        values increase the weight of weighted SSE term.
    - :math:`d_k`, :math:`u_k`: label and features vector at time step k.
    """
    def __init__(self, featuresSize):
        super(RecursiveLeastSquaresFilter, self).__init__()
        self._java_obj = self._new_java_obj("com.github.ozancicek.artan.ml.filter.RecursiveLeastSquaresFilter",
                                            featuresSize, self.uid)
        self._featuresSize = featuresSize

    def setLabelCol(self, value):
        """
        Set label column. Default is "label"

        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(labelCol=value)

    def setFeaturesCol(self, value):
        """
        Set features column. Default is "features"

        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(featuresCol=value)

    def setInitialEstimate(self, value):
        """
        Set initial estimate for model parameters. Default is zero vector.

        Note that if this parameter is set through here, it will result in same initial estimate for all filters.
        For different initial estimates across filters, set the dataframe column for corresponding to initial estimate
        with setInitialEstimateCol.

        :param value: pyspark.ml.linalg.Vector with size (featuresSize)
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(initialState=value)

    def setInitialEstimateCol(self, value):
        """
        Sets the column corresponding to initial estimates
        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(initialStateCol=value)

    def setForgettingFactor(self, value):
        """
        Set forgetting factor, exponentially weights the measurements based on its sequence.

        Default value of 1.0 weights all measurements equally. With smaller values, recent measurements will have
        more weight. Generally set around 0.95 ~ 0.99

        :param value: Float >= 1.0
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(forgettingFactor=value)

    def setRegularizationMatrix(self, value):
        """
        Set regularization matrix governing the influence of the initial estimate (prior). Larger values will
        remove regularization effect, making the filter behave like OLS.

        Default is 10E5 * I

        Note that if this parameter is set through here, it will result in same regularization for all filters.
        For different regularizations across filters, set the dataframe column for corresponding to regularization
        with setRegularizationMatrixCol.

        :param value: pyspark.ml.linalg.Matrix with size (featuresSize, featuresSize)
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(regularizationMatrix=value)

    def setRegularizationMatrixCol(self, value):
        """
        Sets the column corresponding to regularization matrix
        :param value: String
        :return: RecursiveLeastSquaresFilter
         """
        return self._set(regularizationMatrixCol=value)

    def setRegularizationMatrixFactor(self, value):
        """
        Sets the regularization matrix with a float factor, which results in setting the regularization matrix as
        factor * identity

        :param value: Float
        :return: RecursiveLeastSquaresFilter
        """
        regMat = np.eye(self._featuresSize) * value
        rows, cols = regMat.shape
        return self._set(regularizationMatrix=DenseMatrix(rows, cols, regMat.reshape(rows * cols, order="F")))


#!/usr/bin/env python

#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup


VERSION = "0.3.1"

with open("README.md") as f:
    long_description = f.read()

setup(
    name='artan',
    version=VERSION,
    description='Online latent state estimation with Apache Spark.',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='Ozan Cicekci',
    author_email='ozancancicekci@gmail.com',
    url='https://github.com/ozancicek/artan',
    package_data={'': ['LICENSE']},
    include_package_data=True,
    packages=['artan',
              'artan.filter',
              'artan.smoother',
              'artan.state'],
    package_dir={
        'artan': 'python/artan',
        'artan.filter': 'python/artan/filter',
        'artan.smoother': 'python/artan/smoother',
        'artan.state': 'python/artan/state'
    },
    license='http://www.apache.org/licenses/LICENSE-2.0',
    keywords=['pyspark', 'kalman', 'filter', 'sparkml', 'structured', 'streaming'],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6'],
    zip_safe=False)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from pyspark.ml.param import Params, Param, TypeConverters


class HasInitialState(Params):
    """
    Mixin for initial state vector.
    """

    initialState = Param(
        Params._dummy(),
        "initialState", "Initial state vector", typeConverter=TypeConverters.toVector)

    def __init__(self):
        super(HasInitialState, self).__init__()

    def getInitialState(self):
        """
        Gets the value of initial state vector or its default value.
        """
        return self.getOrDefault(self.initialState)


class HasInitialStateCol(Params):
    """
    Mixin for param for initial state column.
    """

    initialStateCol = Param(
        Params._dummy(),
        "initialStateCol",
        "Column name for initial state vector.",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasInitialStateCol, self).__init__()

    def getInitialStateCol(self):
        """
        Gets the value of initial state column or its default value.
        """
        return self.getOrDefault(self.initialStateCol)


class HasInitialCovariance(Params):
    """
    Mixin for param initial covariance matrix.
    """

    initialCovariance = Param(
        Params._dummy(),
        "initialCovariance",
        "Initial covariance matrix",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(HasInitialCovariance, self).__init__()

    def getInitialCovariance(self):
        """
        Gets the value of initial covariance matrix or its default value.
        """
        return self.getOrDefault(self.initialCovariance)


class HasInitialCovarianceCol(Params):
    """
    Mixin for param for initial covariance column.
    """

    initialCovarianceCol = Param(
        Params._dummy(),
        "initialCovarianceCol",
        "Column name for initial covariance vector.",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasInitialCovarianceCol, self).__init__()

    def getInitialCovarianceCol(self):
        """
        Gets the value of initial covariance column or its default value.
        """
        return self.getOrDefault(self.initialCovarianceCol)


class HasProcessModel(Params):
    """
    Mixin for param process model matrix.
    """

    processModel = Param(
        Params._dummy(),
        "processModel",
        "Process model matrix, transitions the state to the next state when applied",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(HasProcessModel, self).__init__()

    def getProcessModel(self):
        """
        Gets the value of process model matrix or its default value.
        """
        return self.getOrDefault(self.processModel)


class HasFadingFactor(Params):
    """
    Mixin for param fading factor.
    """

    fadingFactor = Param(
        Params._dummy(),
        "fadingFactor",
        "Factor controlling the weight of older measurements. With larger factor, more weights will" +
        "be given to recent measurements. Typically, should be really close to 1",
        typeConverter=TypeConverters.toFloat)

    def __init__(self):
        super(HasFadingFactor, self).__init__()

    def getFadingFactor(self):
        """
        Gets the value of fading factor or its default value.
        """
        return self.getOrDefault(self.fadingFactor)


class HasMeasurementModel(Params):
    """
    Mixin for param measurement model matrix.
    """

    measurementModel = Param(
        Params._dummy(),
        "measurementModel",
        "Measurement matrix, when multiplied with the state it should give the measurement vector",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(HasMeasurementModel, self).__init__()

    def getMeasurementModel(self):
        """
        Gets the value of measurement model matrix or its default value.
        """
        return self.getOrDefault(self.measurementModel)


class HasProcessNoise(Params):
    """
    Mixin for param process noise matrix.
    """

    processNoise = Param(
        Params._dummy(),
        "processNoise",
        "Process noise matrix",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(HasProcessNoise, self).__init__()

    def getProcessNoise(self):
        """
        Gets the value of process noise matrix or its default value.
        """
        return self.getOrDefault(self.processNoise)


class HasMeasurementNoise(Params):
    """
    Mixin for param measurement noise matrix.
    """

    measurementNoise = Param(
        Params._dummy(),
        "measurementNoise",
        "Measurement noise matrix",
        typeConverter=TypeConverters.toMatrix)

    def __init__(self):
        super(HasMeasurementNoise, self).__init__()

    def getMeasurementNoise(self):
        """
        Gets the value of measurement noise matrix or its default value.
        """
        return self.getOrDefault(self.measurementNoise)


class HasMeasurementCol(Params):
    """
    Mixin for param for measurement column.
    """

    measurementCol = Param(
        Params._dummy(),
        "measurementCol",
        "Column name for measurement vector. Missing measurements are allowed with nulls in the data",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasMeasurementCol, self).__init__()

    def getMeasurementCol(self):
        """
        Gets the value of measurement column or its default value.
        """
        return self.getOrDefault(self.measurementCol)


class HasMeasurementModelCol(Params):
    """
    Mixin for param for measurement model column.
    """

    measurementModelCol = Param(
        Params._dummy(),
        "measurementModelCol",
        "Column name for specifying measurement model from input DataFrame rather than" +
        "a constant measurement model for all filters",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasMeasurementModelCol, self).__init__()

    def getMeasurementModelCol(self):
        """
        Gets the value of measurement model column or its default value.
        """
        return self.getOrDefault(self.measurementModelCol)


class HasMeasurementNoiseCol(Params):
    """
    Mixin for param for measurement noise column.
    """

    measurementNoiseCol = Param(
        Params._dummy(),
        "measurementNoiseCol",
        "Column name for specifying measurement noise from input DataFrame rather than" +
        "a constant measurement noise for all filters",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasMeasurementNoiseCol, self).__init__()

    def getMeasurementNoiseCol(self):
        """
        Gets the value of measurement noise column or its default value.
        """
        return self.getOrDefault(self.measurementNoiseCol)


class HasProcessModelCol(Params):
    """
    Mixin for param for process model column.
    """

    processModelCol = Param(
        Params._dummy(),
        "processModelCol",
        "Column name for specifying process model from input DataFrame rather than" +
        "a constant measurement model for all filters",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasProcessModelCol, self).__init__()

    def getProcessModelCol(self):
        """
        Gets the value of process model column or its default value.
        """
        return self.getOrDefault(self.processModelCol)


class HasProcessNoiseCol(Params):
    """
    Mixin for param for process noise column.
    """

    processNoiseCol = Param(
        Params._dummy(),
        "processNoiseCol",
        "Column name for specifying process noise from input DataFrame rather than" +
        "a constant measurement noise for all filters",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasProcessNoiseCol, self).__init__()

    def getProcessNoiseCol(self):
        """
        Gets the value of process noise column or its default value.
        """
        return self.getOrDefault(self.processNoiseCol)


class HasControlCol(Params):
    """
    Mixin for param for control column.
    """

    controlCol = Param(
        Params._dummy(),
        "controlCol",
        "Column name for specifying control vector",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasControlCol, self).__init__()

    def getControlCol(self):
        """
        Gets the value of control column or its default value.
        """
        return self.getOrDefault(self.controlCol)


class HasControlFunctionCol(Params):
    """
    Mixin for param for control function column.
    """

    controlFunctionCol = Param(
        Params._dummy(),
        "controlFunctionCol",
        "Column name for specifying control matrix",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasControlFunctionCol, self).__init__()

    def getControlFunctionCol(self):
        """
        Gets the value of control function column or its default value.
        """
        return self.getOrDefault(self.controlFunctionCol)


class HasCalculateMahalanobis(Params):
    """
    Mixin for param for enabling mahalanobis calculation.
    """

    calculateMahalanobis = Param(
        Params._dummy(),
        "calculateMahalanobis",
        "When true, mahalanobis distance of residual will be calculated & added to output DataFrame." +
        "Default is false.",
        typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(HasCalculateMahalanobis, self).__init__()

    def getCalculateMahalanobis(self):
        """
        Gets the value of mahalanobis calcuation flag.
        """
        return self.getOrDefault(self.calculateMahalanobis)


class HasCalculateLoglikelihood(Params):
    """
    Mixin for param for enabling loglikelihood calculation.
    """

    calculateLoglikelihood = Param(
        Params._dummy(),
        "calculateLoglikelihood",
        "When true, loglikelihood of residual will be calculated & added to output DataFrame. Default is false",
        typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(HasCalculateLoglikelihood, self).__init__()

    def getCalculateLoglikelihood(self):
        """
        Gets the value of loglikelihood calculation flag.
        """
        return self.getOrDefault(self.calculateLoglikelihood)


class HasOutputSystemMatrices(Params):
    """
    Mixin for param for enabling the output of system matrices along with the state.
    """

    outputSystemMatrices = Param(
        Params._dummy(),
        "outputSystemMatrices",
        "When true, the system matrices will be added to output DataFrame. Default is false",
        typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(HasOutputSystemMatrices, self).__init__()

    def getOutputSystemMatrices(self):
        """
        Gets the value of loglikelihood calculation flag.
        """
        return self.getOrDefault(self.outputSystemMatrices)


class HasCalculateSlidingLikelihood(Params):
    """
    Mixin param for enabling sliding likelihood calculation
    """
    calculateSlidingLikelihood = Param(
        Params._dummy(),
        "calculateSlidingLikelihood",
        "When true, sliding likelihood sum of residual will be calculated & added to output DataFrame. Default is false",
        typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(HasCalculateSlidingLikelihood, self).__init__()

    def getCalculateSlidingLikelihood(self):
        """
        Gets the value of sliding likelihood calculation flag
        """
        return self.getOrDefault(self.calculateSlidingLikelihood)


class HasSlidingLikelihoodWindow(Params):
    """
    Mixin param for sliding likelihood window duration
    """
    slidingLikelihoodWindow = Param(
        Params._dummy(),
        "slidingLikelihoodWindow",
        "Number of consecutive measurements to include in the total likelihood calculation",
        typeConverter=TypeConverters.toInt)

    def __init__(self):
        super(HasSlidingLikelihoodWindow, self).__init__()

    def getSlidingLikelihoodWindow(self):
        """
        Gets the value of sliding likelihood window
        """
        return self.getOrDefault(self.slidingLikelihoodWindow)


class HasMultipleModelMeasurementWindowDuration(Params):

    multipleModelMeasurementWindowDuration = Param(
        Params._dummy(),
        "multipleModelMeasurementWindowDuration",
        "Window duration for grouping measurements in same window for MMAE filter aggregation",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasMultipleModelMeasurementWindowDuration, self).__init__()

    def getMultipleModelMeasurementWindowDuration(self):
        """
        Gets the value of mmae measureent window duration
        """
        return self.getOrDefault(self.multipleModelMeasurementWindowDuration)


class HasMultipleModelAdaptiveEstimationEnabled(Params):

    multipleModelAdaptiveEstimationEnabled = Param(
        Params._dummy(),
        "multipleModelAdaptiveEstimationEnabled",
        "Flag for enabling  Multiple Model Adaptive Estimation (MMAE) output mode. When enabled," + "" +
        "MMAE mode outputs a single state estimate from the output of all kalman states of the transformer." +
        "States are weighted based on their sliding likelihood",
        typeConverter=TypeConverters.toBoolean)

    def __init__(self):
        super(HasMultipleModelAdaptiveEstimationEnabled, self).__init__()

    def getMultipleModelAdaptiveEstimationEnabled(self):
        """
        Gets the value of MMAE output mode flag
        """
        return self.getOrDefault(self.multipleModelAdaptiveEstimationEnabled)


class KalmanFilterParams(HasInitialState, HasInitialCovariance, HasInitialStateCol,
                         HasInitialCovarianceCol, HasProcessModel, HasFadingFactor, HasMeasurementModel,
                         HasMeasurementNoise, HasProcessNoise, HasMeasurementCol,
                         HasMeasurementModelCol, HasMeasurementNoiseCol, HasProcessModelCol,
                         HasProcessNoiseCol, HasControlCol, HasControlFunctionCol,
                         HasCalculateMahalanobis, HasCalculateLoglikelihood,
                         HasOutputSystemMatrices, HasCalculateSlidingLikelihood,
                         HasSlidingLikelihoodWindow):
    """
    Mixin for kalman filter parameters
    """
    def setInitialState(self, value):
        """
        Set the initial state vector with size (stateSize).

        It will be applied to all states. If the state timeouts and starts receiving
        measurements after timeout, it will again start from this initial state vector. Default is zero.

        Note that if this parameter is set through here, it will result in same initial state for all filters.
        For different initial states across filters, set the dataframe column for corresponding to initial state
        with setInitialStateCol.

        :param value: pyspark.ml.linalg.Vector with size (stateSize)
        :return: KalmanFilter
        """
        return self._set(initialState=value)

    def setInitialStateCol(self, value):
        """
        Set the column corresponding to initial state vector. Overrides setInitialState setting.

        :param value: String
        :return: KalmanFilter
        """
        return self._set(initialStateCol=value)

    def setInitialCovariance(self, value):
        """
        Set the initial covariance matrix with dimensions (stateSize, stateSize)

        It will be applied to all states. If the state timeouts and starts receiving
        measurements after timeout, it will again start from this initial covariance vector. Default is identity matrix.
        :param value: pyspark.ml.linalg.Matrix with dimensions (stateSize, stateSize)
        :return: KalmanFilter
        """
        return self._set(initialCovariance=value)

    def setInitialCovarianceCol(self, value):
        """
        Set the column corresponding to initial covariance matrix. Overrides setInitialCovariance setting.

        :param value: String
        :return: KalmanFilter
        """
        return self._set(initialCovarianceCol=value)

    def setFadingFactor(self, value):
        """
        Fading factor for giving more weights to more recent measurements. If needed, it should be greater than one.
        Typically set around 1.01 ~ 1.05. Default is 1.0, which will result in equally weighted measurements.

        :param value: Float >= 1.0
        :return: KalmanFilter
        """
        return self._set(fadingFactor=value)

    def setProcessModel(self, value):
        """
        Set default value for process model matrix with dimensions (stateSize, stateSize) which governs
        state transition.

        Note that if this parameter is set through here, it will result in same process model for all filters &
        measurements. For different process models across filters or measurements, set a dataframe column for process
        model from setProcessModelCol.

        Default is identity matrix.

        :param value: pyspark.ml.linalg.Matrix with dimensions (stateSize, stateSize)
        :return: KalmanFilter
        """
        return self._set(processModel=value)

    def setProcessNoise(self, value):
        """
        Set default value for process noise matrix with dimensions (stateSize, stateSize).

        Note that if this parameter is set through here, it will result in same process noise for all filters &
        measurements. For different process noise values across filters or measurements, set a dataframe column
        for process noise from setProcessNoiseCol.

        Default is identity matrix.

        :param value: pyspark.ml.linalg.Matrix with dimensions (stateSize, StateSize)
        :return: KalmanFilter
        """
        return self._set(processNoise=value)

    def setMeasurementModel(self, value):
        """
        Set default value for measurement model matrix with dimensions (stateSize, measurementSize)
        which maps states to measurement.

        Note that if this parameter is set through here, it will result in same measurement model for all filters &
        measurements. For different measurement models across filters or measurements, set a dataframe column for
        measurement model from setMeasurementModelCol.

        Default value maps the first state value to measurements.

        :param value: pyspark.ml.linalg.Matrix with dimensions (stateSize, measurementSize)
        :return: KalmanFilter
        """
        return self._set(measurementModel=value)

    def setMeasurementNoise(self, value):
        """
        Set default value for measurement noise matrix with dimensions (measurementSize, measurementSize).

        Note that if this parameter is set through here, it will result in same measurement noise for all filters &
        measurements. For different measurement noise values across filters or measurements,
        set a dataframe column for measurement noise from setMeasurementNoiseCol.

        Default is identity matrix.

        :param value: pyspark.ml.linalg.Matrix with dimensions (measurementSize, measurementSize)
        :return: KalmanFilter
        """
        return self._set(measurementNoise=value)

    def setMeasurementCol(self, value):
        """
        Set the column corresponding to measurements.

        The vectors in the column should be of size (measurementSize). null values are allowed,
        which will result in only state prediction step.

        :param value: pyspark.ml.linalg.Vector with size measurementSize
        :return: KalmanFilter
        """
        return self._set(measurementCol=value)

    def setMeasurementModelCol(self, value):
        """
        Set the column for input measurement model matrices

        Measurement model matrices should have dimensions (stateSize, measurementSize)

        :param value: String
        :return: KalmanFilter
        """
        return self._set(measurementModelCol=value)

    def setProcessModelCol(self, value):
        """
        Set the column for input process model matrices.

        Process model matrices should have dimensions (stateSize, stateSize)

        :param value: String
        :return: KalmanFilter
        """
        return self._set(processModelCol=value)

    def setProcessNoiseCol(self, value):
        """
        Set the column for input process noise matrices.

        Process noise matrices should have dimensions (stateSize, stateSize)

        :param value: String
        :return: KalmanFilter
        """
        return self._set(processNoiseCol=value)

    def setMeasurementNoiseCol(self, value):
        """
        Set the column for input measurement noise matrices.

        Measurement noise matrices should have dimensions (measurementSize, measurementSize)

        :param value: String
        :return: KalmanFilter
        """
        return self._set(measurementNoiseCol=value)

    def setControlCol(self, value):
        """
        Set the column for input control vectors.

        Control vectors should have compatible size with control function (controlVectorSize). The product of
        control matrix & vector should produce a vector with stateSize. null values are allowed,
        which will result in state transition without control input.

        :param value: String
        :return: KalmanFilter
        """
        return self._set(controlCol=value)

    def setControlFunctionCol(self, value):
        """
        Set the column for input control matrices.

        Control matrices should have dimensions (stateSize, controlVectorSize). null values are allowed, which will
        result in state transition without control input

        :param value: String
        :return: KalmanFilter
        """
        return self._set(controlFunctionCol=value)

    def setCalculateLogLikelihood(self):
        """
        Optionally calculate loglikelihood of each measurement & add it to output dataframe. Loglikelihood is calculated
        from residual vector & residual covariance matrix.

        Not enabled by default.

        :return: KalmanFilter
        """
        return self._set(calculateLoglikelihood=True)

    def setCalculateMahalanobis(self):
        """
        Optionally calculate mahalanobis distance metric of each measurement & add it to output dataframe.
        Mahalanobis distance is calculated from residual vector & residual covariance matrix.

        Not enabled by default.

        :return: KalmanFilter
        """
        return self._set(calculateMahalanobis=True)

    def setOutputSystemMatrices(self):
        """
        Optionally add system matrices to output dataframe returned by the transformer.

        Default is false

        :return: KalmanFilter
        """
        return self._set(outputSystemMatrices=True)

    def setCalculateSlidingLikelihood(self):
        """
        Optionally calculate a sliding likelihood across consecutive measurements

        Default is false

        :return: KalmanFilter
        """
        return self._set(calculateSlidingLikelihood=True)

    def setSlidingLikelihoodWindow(self, value):
        """
        Set the param for number of consecutive measurements to include in the total likelihood calculation

        Default is 1

        :param value: Integer
        :return: KalmanFilter
        """
        return self._set(slidingLikelihoodWindow=value) \
            .setCalculateSlidingLikelihood()


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from pyspark.ml.common import inherit_doc
from pyspark.ml.param import Params, Param, TypeConverters

from artan.state import StatefulTransformer
from artan.filter.filter_params import KalmanFilterParams


class HasFixedLag(Params):
    """
    Mixin for param for fixed lag
    """

    fixedLag = Param(
        Params._dummy(),
        "fixedLag", "Fixed lag", typeConverter=TypeConverters.toInt)

    def __init__(self):
        super(HasFixedLag, self).__init__()

    def getFixedLag(self):
        """
        Gets the value of fixed lag or its default value.
        """
        return self.getOrDefault(self.fixedLag)


@inherit_doc
class LinearKalmanSmoother(StatefulTransformer, KalmanFilterParams, HasFixedLag):
    """
    Fixed lag linear kalman smoother using Rauch-Tung-Striebel method. The smoother is implemented with a
    stateful spark transformer for running parallel smoother /w spark dataframes.
    Transforms an input dataframe of noisy measurements to dataframe of state estimates using
    stateful spark transformations, which can be used in both streaming and batch applications.

    At a time step k and a fixed lag N, the fixed lag linear kalman smoother computes the state estimates of a linear
    kalman filter based on all measurements made between step k and step k-t. For each time step k >= N, the smoother
    outputs an estimate for all the time steps between k and k-N. When k < N, the smoother doesn't output any estimates.
    As a result, the memory requirements of this filter is N times of a linear kalman filter. Since the smoother
    outputs multiple estimates for a single measurement, it is advised to set event time column
    of the measurements with setEventTimeCol.
    """
    def __init__(self, stateSize, measurementSize):
        super(LinearKalmanSmoother, self).__init__()
        self._java_obj = self._new_java_obj("com.github.ozancicek.artan.ml.smoother.LinearKalmanSmoother",
                                            stateSize, measurementSize, self.uid)

    def setFixedLag(self, value):
        """
        Sets the smoother fixed lag

        Default is 2.
        :param value: Int
        :return: LinearKalmanSmoother
        """
        return self._set(fixedLag=value)


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#


from pyspark.ml.param import Params, Param, TypeConverters
from pyspark.ml.wrapper import JavaTransformer
from pyspark.ml.common import inherit_doc


class HasStateKeyCol(Params):
    """
    Mixin for param for state key column.
    """

    stateKeyCol = Param(
        Params._dummy(),
        "stateKeyCol",
        "State key column. State keys uniquely identify the each state in stateful transformers," +
        "thus controlling the number of states and the degree of parallelization",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasStateKeyCol, self).__init__()

    def getStateKeyCol(self):
        """
        Gets the value of state key column or its default value.
        """
        return self.getOrDefault(self.stateKeyCol)


class HasEventTimeCol(Params):
    """
    Mixin for param for event time column.
    """

    eventTimeCol = Param(
        Params._dummy(),
        "eventTimeCol",
        "Column marking the event time of the received measurements",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasEventTimeCol, self).__init__()

    def getEventTimeCol(self):
        """
        Gets the value of event time column or its default value.
        """
        return self.getOrDefault(self.eventTimeCol)


class HasWatermarkDuration(Params):
    """
    Mixin for param for watermark duration.
    """

    watermarkDuration = Param(
        Params._dummy(),
        "watermarkDuration",
        "Watermark duration",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasWatermarkDuration, self).__init__()

    def getWatermarkDuration(self):
        """
        Gets the value of watermark duration or its default value.
        """
        return self.getOrDefault(self.watermarkDuration)


class HasStateTimeoutMode(Params):
    """
    Mixin for param for state timeout mode for clearing states without updates, one of "none", "process" or "event".
    """

    timeoutMode = Param(
        Params._dummy(),
        "timeoutMode",
        "Timeout mode for clearing the states that didn't receive measurements.",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasStateTimeoutMode, self).__init__()

    def getTimeoutMode(self):
        """
        Gets the value of timeout mode or its default value.
        """
        return self.getOrDefault(self.timeoutMode)


class HasStateTimeoutDuration(Params):
    """
    Mixin for param for state timeout duration.
    """

    stateTimeoutDuration = Param(
        Params._dummy(),
        "stateTimeoutDuration",
        "Duration to wait before timing out the state",
        typeConverter=TypeConverters.toString)

    def __init__(self):
        super(HasStateTimeoutDuration, self).__init__()

    def getStateTimeoutDuration(self):
        """
        Gets the value of state timeout duration or its default value.
        """
        return self.getOrDefault(self.stateTimeoutDuration)


@inherit_doc
class StatefulTransformer(JavaTransformer,
                          HasStateKeyCol, HasEventTimeCol, HasWatermarkDuration,
                          HasStateTimeoutDuration, HasStateTimeoutMode):
    """
    Base mixin for stateful transformations
    """
    def setStateKeyCol(self, value):
        """
        Sets the state key column. Each value in the column should uniquely identify a stateful transformer. Each
        unique value will result in a separate state.
        """
        return self._set(stateKeyCol=value)

    def setStateTimeoutMode(self, value):
        """
        Sets the state timeout mode. Supported values are 'none', 'process' and 'event'. Enabling state timeout will
        clear the state after a certain timeout duration which can be set. If a state receives measurements after
        it times out, the state will be initialized as if it received no measurements.

        - 'none': No state timeout, state is kept indefinitely.

        - 'process': Process time based state timeout, state will be cleared if no measurements are received for
            a duration based on processing time. Effects all states. Timeout duration must be set with
            setStateTimeoutDuration.

        - 'event': Event time based state timeout, state will be cleared if no measurements are recieved for a duration
            based on event time determined by watermark. Effects all states. Timeout duration must be set with
            setStateTimeoutDuration. Additionally, event time column and it's watermark duration must be set with
            setEventTimeCol and setWatermarkDuration. Note that this will result in dropping measurements occuring later
            than the watermark.

        Default is 'none'

        :param value: String, one of 'none', 'process' or 'event'
        :return: StatefulTransformer
        """
        return self._set(timeoutMode=value)

    def setEventTimeCol(self, value):
        """
        Sets the event time column in the input DataFrame.

        :param value: String, column name of the eventTime timestamp
        :return: StatefulTransformer
        """
        return self._set(eventTimeCol=value)

    def setWatermarkDuration(self, value):
        """
        Set the watermark duration for all states, only valid when state timeout mode is 'event'.
        Must be a valid duration string, such as '10 minutes'.

        :param value: String, duration specifying watermark from eventTime timestamp
        :return: StatefulTransformer
        """
        return self._set(watermarkDuration=value)


    def setStateTimeoutDuration(self, value):
        """
        Sets the state timeout duration for all states, only valid when state timeout mode is not 'none'.

        Must be a valid duration string, such as '10 minutes'.

        :param value: String, duration specifying state timeout
        :return: StatefulTransformer
        """
        return self._set(stateTimeoutDuration=value)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at

#  http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from pyspark.ml.param import Params, Param, TypeConverters
from pyspark.ml.common import inherit_doc
from pyspark.ml.param.shared import HasLabelCol, HasFeaturesCol
from artan.state import StatefulTransformer
from artan.filter.filter_params import HasInitialState


class _HasLearningRate(Params):
    """
    Mixin for param Normalized LMS learning rate
    """

    learningRate = Param(
        Params._dummy(),
        "learningRate",
        "Learning rate for Normalized LMS. If there is no interference, the default value of 1.0 is optimal",
        typeConverter=TypeConverters.toFloat)

    def __init__(self):
        super(_HasLearningRate, self).__init__()

    def getLearningRate(self):
        """
        Gets the value of learning rate or its default value.
        """
        return self.getOrDefault(self.learningRate)


class _HasRegularizationConstant(Params):
    """
    Mixin for param for regularization constant.
    """

    regularizationConstant = Param(
        Params._dummy(),
        "regularizationConstant",
        "Regularization term for stability, default is 1.0",
        typeConverter=TypeConverters.toFloat)

    def __init__(self):
        super(_HasRegularizationConstant, self).__init__()

    def getRegularizationConstant(self):
        """
        Gets the value of regularization constant or its default value.
        """
        return self.getOrDefault(self.regularizationConstant)


@inherit_doc
class LeastMeanSquaresFilter(StatefulTransformer, HasInitialState,
                             _HasLearningRate, _HasRegularizationConstant,
                             HasLabelCol, HasFeaturesCol):
    """
    Normalized Least Mean Squares filter, implemented with a stateful spark Transformer for running parallel
    filters /w spark dataframes. Transforms an input dataframe of observations to a dataframe of model parameters
    using stateful spark transformations, which can be used in both streaming and batch applications.

    Belonging to stochastic gradient descent type of methods, LMS minimizes SSE on each measurement
    based on the expectation of steepest descending gradient.

    Let :math:`w` denote the model parameter vector, :math:`u` denote the features vector,
    and :math:`d` the label corresponding to :math:`u`. Normalized LMS computes :math:`w` at step :math:`k`
    recursively by:

    .. math ::

        e &= d - u \cdot w_{k-1} \\

        w_k &= w_{k-1} + m * e * u * (c + u \cdot u)^{-1}


    Where
    - :math:`m`: Learning rate
    - :math:`c`: Regularization constant
    """
    def __init__(self, featuresSize):
        super(LeastMeanSquaresFilter, self).__init__()
        self._java_obj = self._new_java_obj("com.github.ozancicek.artan.ml.filter.LeastMeanSquaresFilter",
                                            featuresSize, self.uid)
        self._featuresSize = featuresSize

    def setLabelCol(self, value):
        """
        Set label column. Default is "label"

        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(labelCol=value)

    def setFeaturesCol(self, value):
        """
        Set features column. Default is "features"

        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(featuresCol=value)

    def setInitialEstimate(self, value):
        """
        Set initial estimate for model parameters. Default is zero vector.

        Note that if this parameter is set through here, it will result in same initial estimate for all filters.
        For different initial estimates across filters, set the dataframe column for corresponding to initial estimate
        with setInitialEstimateCol.

        :param value: pyspark.ml.linalg.Vector with size (featuresSize)
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(initialState=value)

    def setInitialEstimateCol(self, value):
        """
        Sets the column corresponding to initial estimates
        :param value: String
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(initialStateCol=value)

    def setLearningRate(self, value):
        """
        Set learning rate controlling the speed of convergence. Without noise, 1.0 is optimal.

        Default is 1.0

        :param value: Float
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(learningRate=value)

    def setRegularizationConstant(self, value):
        """
        Set constant for regularization controlling stability. Larger values increase stability but degrade
        convergence performance. Generally set to a small constant.

        Default is 1.0

        :param value: Float
        :return: RecursiveLeastSquaresFilter
        """
        return self._set(regularizationConstant=value)